// solely an AP with the facilities of updating itself; by .../update or by an assignment of the url. version 20210603.0.1.
#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <ESP8266HTTPUpdateServer.h>
#include <ESP8266httpUpdate.h>

#define APSSID "ESP_TESTAP"
#define APPSK  "87654321"

const char *ssid = APSSID;
const char *password = APPSK;

ESP8266WebServer server(80);
ESP8266HTTPUpdateServer httpUpdater; // webupdater, will be http://HOSTNAME.local/update.
String webupd_addr="", uri_list="";
unsigned force_update=0;

void setup() {
  Serial.begin(115200);
  while (!Serial) delay(50);
  Serial.println();
  Serial.print("Configuring access point...");
  WiFi.softAP(ssid, password);
  IPAddress myIP = WiFi.softAPIP();
  Serial.print("AP IP address: ");
  Serial.println(myIP);

    server.on("/", []() {server.send(200, "text/html", "<h1>WiFi AP ESP_TESTAP</h1>");});

    server.on("/version", []() {server.send(200, "text/plain", "[20210603.0.1]");});

    server.on("/reset", []() {
        server.send(200, "text/plain", "now reset...");
        delay(1000);
        while (1);
    });

    server.on("/s", []() {server.send(200, "text/plain", String(random(10)+1).c_str());});

    // assign the update website url to this device(it is intended for this local device connecting to)
    server.on("/webupdinfo", []() { // http://<ip address>/webupdinfo?addr=<remote url; e.g., http://192.168.1.1/a.bin>.
        String addr=server.arg("addr");
        if (addr=="") server.send(404, "text/plain", "0. page not found.");
        else {
            webupd_addr=addr;
            server.send(200, "text/plain", ("1. accepted: "+webupd_addr).c_str());
        }
    });

    // force global update immediately
    server.on("/forceupdate", []() {
        force_update=1;
        server.send(200, "text/plain", String("1. accepted. on 10 seconds to start updating...").c_str());
    });

    server.on("/listcmds", []() { // list all commands.
        server.send(200, "text/plain", uri_list.c_str());
    });

    // use uri_list+=str; to expand uri list.
    uri_list=\
        "/listcmds\r\n"\
        "/\r\n"\
        "/version\r\n"\
        "/reset\r\n"\
        "/s\r\n"\
        "/update (by selecting an update file from pc and upload to update)\r\n"\
        "/webupdinfo?addr=<remote url; e.g., http://192.168.1.1/a.bin>\r\n"\
        "/forceupdate\r\n";
        "/s\r\n";

  httpUpdater.setup(&server);
  server.begin();
  Serial.println("HTTP server started");
}

bool Timeout(unsigned ms, unsigned &store_start_time_us, bool one_shot=false, bool one_shot_lasting=false){
    // note: the var store_start_time_us must be exclusively used by this function because it could be modified by this function.
    // !!!!
    // especially note that the micros() will wrap around after 4294967296us, which is around 70 minutes.
    // so, do not use timeout value > 35 minutes; or another abnormal case is past time > 35 minutes before checking.
    // that is, avoid from abnormal of the above two is must by user.

    if (one_shot && !store_start_time_us) return one_shot_lasting;

    ///// if (micros()>=(ms*1000+store_start_time_us)){ ///// this is not ok, will abnormal after power-on 70 minutes.
    if (int(micros())-int(ms*1000+store_start_time_us)>=0){ ///// this is ok

        store_start_time_us=one_shot? 0: micros();
        return true;
    }
    return false;
}

// -------------------------- Remote updater ------------------------------
// the device can not update itself because this updater has yet not found any exit point.
void gbl_update_started() {
  Serial.println("CALLBACK:  HTTP update process started");
}

void gbl_update_finished() {
  Serial.println("CALLBACK:  HTTP update process finished");
}

void gbl_update_progress(int cur, int total) {
  Serial.printf("CALLBACK:  HTTP update process at %d of %d bytes...\n", cur, total);
}

void gbl_update_error(int err) {
  Serial.printf("CALLBACK:  HTTP update fatal error code %d\n", err);
}

void gbl_updater(const char *url) {
    // wait for WiFi connection

    WiFiClient client;

    // The line below is optional. It can be used to blink the LED on the board during flashing
    // The LED will be on during download of one buffer of data from the network. The LED will
    // be off during writing that buffer to flash
    // On a good connection the LED should flash regularly. On a bad connection the LED will be
    // on much longer than it will be off. Other pins than LED_BUILTIN may be used. The second
    // value is used to put the LED on. If the LED is on with HIGH, that value should be passed
    // ESPhttpUpdate.setLedPin(LED_BUILTIN, LOW);

    // Add optional callback notifiers
    ESPhttpUpdate.onStart(gbl_update_started);
    ESPhttpUpdate.onEnd(gbl_update_finished);
    ESPhttpUpdate.onProgress(gbl_update_progress);
    ESPhttpUpdate.onError(gbl_update_error);

    t_httpUpdate_return ret = ESPhttpUpdate.update(client, url);
    // Or:
    // t_httpUpdate_return ret = ESPhttpUpdate.update(client, "server", 80, "file.bin");

    switch (ret) {
        case HTTP_UPDATE_FAILED:
            Serial.printf("HTTP_UPDATE_FAILD Error (%d): %s\r\n", ESPhttpUpdate.getLastError(), ESPhttpUpdate.getLastErrorString().c_str());
            break;
        case HTTP_UPDATE_NO_UPDATES:
            Serial.println("HTTP_UPDATE_NO_UPDATES");
            break;
        case HTTP_UPDATE_OK:
            Serial.println("HTTP_UPDATE_OK");
            break;
    }
}

void loop() {
    server.handleClient();
    if (force_update && webupd_addr!=""){
        static unsigned tmo1=micros();
        if (force_update==1){
            force_update=2;
            tmo1=micros();
        }
        else if ((force_update==2) && Timeout(10000, tmo1, true)){ // 10 seconds to start
            force_update=0;
            gbl_updater(webupd_addr.c_str());
        }
    }
}
